import json
import os
import time
import warnings
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from threading import Lock
from typing import Dict, Generator, List, Optional, Tuple, Union

import requests

from lagent.schema import ModelStatusCode
from lagent.utils.util import filter_suffix
from .base_api import BaseAPILLM

warnings.simplefilter('default')

SENSENOVA_API_BASE = 'https://api.sensenova.cn/v1/llm/chat-completions'

sensechat_models = {'SenseChat-5': 131072, 'SenseChat-5-Cantonese': 32768}


class SensenovaAPI(BaseAPILLM):
    """Model wrapper around SenseTime's models.

    Args:
        model_type (str): The name of SenseTime's model.
        retry (int): Number of retires if the API call fails. Defaults to 2.
        key (str or List[str]): SenseTime key(s). In particular, when it
            is set to "ENV", the key will be fetched from the environment
            variable $SENSENOVA_API_KEY. If it's a list, the keys will be
            used in round-robin manner. Defaults to 'ENV'.
        meta_template (Dict, optional): The model's meta prompt
            template if needed, in case the requirement of injecting or
            wrapping of any meta instructions.
        sensenova_api_base (str): The base url of SenseTime's API. Defaults to
            'https://api.sensenova.cn/v1/llm/chat-completions'.
        gen_params: Default generation configuration which could be overridden
            on the fly of generation.
    """

    is_api: bool = True

    def __init__(
        self,
        model_type: str = 'SenseChat-5-Cantonese',
        retry: int = 2,
        json_mode: bool = False,
        key: Union[str, List[str]] = 'ENV',
        meta_template: Optional[Dict] = [
            dict(role='system', api_role='system'),
            dict(role='user', api_role='user'),
            dict(role='assistant', api_role='assistant'),
            dict(role='environment', api_role='system'),
        ],
        sensenova_api_base: str = SENSENOVA_API_BASE,
        proxies: Optional[Dict] = None,
        **gen_params,
    ):

        super().__init__(
            model_type=model_type,
            meta_template=meta_template,
            retry=retry,
            **gen_params,
        )
        self.logger = getLogger(__name__)

        if isinstance(key, str):
            # First, apply for SenseNova's ak and sk from SenseTime staff
            # Then, generated SENSENOVA_API_KEY using lagent.utils.gen_key.auto_gen_jwt_token(ak, sk)
            self.keys = [
                os.getenv('SENSENOVA_API_KEY') if key == 'ENV' else key
            ]
        else:
            self.keys = key

        # record invalid keys and skip them when requesting API
        # - keys have insufficient_quota
        self.invalid_keys = set()

        self.key_ctr = 0
        self.url = sensenova_api_base
        self.model_type = model_type
        self.proxies = proxies
        self.json_mode = json_mode

    def chat(
        self,
        inputs: Union[List[dict], List[List[dict]]],
        **gen_params,
    ) -> Union[str, List[str]]:
        """Generate responses given the contexts.

        Args:
            inputs (Union[List[dict], List[List[dict]]]): a list of messages
                or list of lists of messages
            gen_params: additional generation configuration

        Returns:
            Union[str, List[str]]: generated string(s)
        """
        assert isinstance(inputs, list)
        if 'max_tokens' in gen_params:
            raise NotImplementedError('unsupported parameter: max_tokens')
        gen_params = {**self.gen_params, **gen_params}
        with ThreadPoolExecutor(max_workers=20) as executor:
            tasks = [
                executor.submit(self._chat,
                                self.template_parser._prompt2api(messages),
                                **gen_params)
                for messages in (
                    [inputs] if isinstance(inputs[0], dict) else inputs)
            ]
        ret = [task.result() for task in tasks]
        return ret[0] if isinstance(inputs[0], dict) else ret

    def stream_chat(
        self,
        inputs: List[dict],
        **gen_params,
    ) -> Generator[Tuple[ModelStatusCode, str, Optional[str]], None, None]:
        """Generate responses given the contexts.

        Args:
            inputs (List[dict]): a list of messages
            gen_params: additional generation configuration

        Yields:
            Tuple[ModelStatusCode, str, Optional[str]]: Status code, generated string, and optional metadata
        """
        assert isinstance(inputs, list)
        if 'max_tokens' in gen_params:
            raise NotImplementedError('unsupported parameter: max_tokens')
        gen_params = self.update_gen_params(**gen_params)
        gen_params['stream'] = True

        resp = ''
        finished = False
        stop_words = gen_params.get('stop_words') or []
        messages = self.template_parser._prompt2api(inputs)
        for text in self._stream_chat(messages, **gen_params):
            # TODO 测试 resp = text 还是 resp += text
            resp += text
            if not resp:
                continue
            # remove stop_words
            for sw in stop_words:
                if sw in resp:
                    resp = filter_suffix(resp, stop_words)
                    finished = True
                    break
            yield ModelStatusCode.STREAM_ING, resp, None
            if finished:
                break
        yield ModelStatusCode.END, resp, None

    def _chat(self, messages: List[dict], **gen_params) -> str:
        """Generate completion from a list of templates.

        Args:
            messages (List[dict]): a list of prompt dictionaries
            gen_params: additional generation configuration

        Returns:
            str: The generated string.
        """
        assert isinstance(messages, list)

        header, data = self.generate_request_data(
            model_type=self.model_type,
            messages=messages,
            gen_params=gen_params,
            json_mode=self.json_mode,
        )

        max_num_retries = 0
        while max_num_retries < self.retry:
            self._wait()

            with Lock():
                if len(self.invalid_keys) == len(self.keys):
                    raise RuntimeError('All keys have insufficient quota.')

                # find the next valid key
                while True:
                    self.key_ctr += 1
                    if self.key_ctr == len(self.keys):
                        self.key_ctr = 0

                    if self.keys[self.key_ctr] not in self.invalid_keys:
                        break

                key = self.keys[self.key_ctr]
                header['Authorization'] = f'Bearer {key}'

            response = dict()
            try:
                raw_response = requests.post(
                    self.url,
                    headers=header,
                    data=json.dumps(data),
                    proxies=self.proxies,
                )
                response = raw_response.json()
                return response['choices'][0]['message']['content'].strip()
            except requests.ConnectionError:
                print('Got connection error, retrying...')
                continue
            except requests.JSONDecodeError:
                print('JsonDecode error, got', str(raw_response.content))
                continue
            except KeyError:
                if 'error' in response:
                    if response['error']['code'] == 'rate_limit_exceeded':
                        time.sleep(1)
                        continue
                    elif response['error']['code'] == 'insufficient_quota':
                        self.invalid_keys.add(key)
                        self.logger.warn(f'insufficient_quota key: {key}')
                        continue

                    print('Find error message in response: ',
                          str(response['error']))
            except Exception as error:
                print(str(error))
            max_num_retries += 1

        raise RuntimeError('Calling SenseTime failed after retrying for '
                           f'{max_num_retries} times. Check the logs for '
                           'details.')

    def _stream_chat(self, messages: List[dict], **gen_params) -> str:
        """Generate completion from a list of templates.

        Args:
            messages (List[dict]): a list of prompt dictionaries
            gen_params: additional generation configuration

        Returns:
            str: The generated string.
        """

        def streaming(raw_response):
            for chunk in raw_response.iter_lines():
                if chunk:
                    try:
                        decoded_chunk = chunk.decode('utf-8')
                        # print(f"Decoded chunk: {decoded_chunk}")

                        if decoded_chunk == 'data:[DONE]':
                            # print("Stream ended")
                            break

                        if decoded_chunk.startswith('data:'):
                            json_str = decoded_chunk[5:]
                            chunk_data = json.loads(json_str)

                            if 'data' in chunk_data and 'choices' in chunk_data[
                                    'data']:
                                choice = chunk_data['data']['choices'][0]
                                if 'delta' in choice:
                                    content = choice['delta']
                                    yield content
                        else:
                            print(f'Unexpected format: {decoded_chunk}')

                    except json.JSONDecodeError as e:
                        print(f'JSON parsing error: {e}')
                    except Exception as e:
                        print(
                            f'An error occurred while processing the chunk: {e}'
                        )

        assert isinstance(messages, list)

        header, data = self.generate_request_data(
            model_type=self.model_type,
            messages=messages,
            gen_params=gen_params,
            json_mode=self.json_mode,
        )

        max_num_retries = 0
        while max_num_retries < self.retry:
            if len(self.invalid_keys) == len(self.keys):
                raise RuntimeError('All keys have insufficient quota.')

            # find the next valid key
            while True:
                self.key_ctr += 1
                if self.key_ctr == len(self.keys):
                    self.key_ctr = 0

                if self.keys[self.key_ctr] not in self.invalid_keys:
                    break

            key = self.keys[self.key_ctr]
            header['Authorization'] = f'Bearer {key}'

            response = dict()
            try:
                raw_response = requests.post(
                    self.url,
                    headers=header,
                    data=json.dumps(data),
                    proxies=self.proxies,
                )
                return streaming(raw_response)
            except requests.ConnectionError:
                print('Got connection error, retrying...')
                continue
            except requests.JSONDecodeError:
                print('JsonDecode error, got', str(raw_response.content))
                continue
            except KeyError:
                if 'error' in response:
                    if response['error']['code'] == 'rate_limit_exceeded':
                        time.sleep(1)
                        continue
                    elif response['error']['code'] == 'insufficient_quota':
                        self.invalid_keys.add(key)
                        self.logger.warn(f'insufficient_quota key: {key}')
                        continue

                    print('Find error message in response: ',
                          str(response['error']))
            except Exception as error:
                print(str(error))
            max_num_retries += 1

        raise RuntimeError('Calling SenseTime failed after retrying for '
                           f'{max_num_retries} times. Check the logs for '
                           'details.')

    def generate_request_data(self,
                              model_type,
                              messages,
                              gen_params,
                              json_mode=False):
        """
        Generates the request data for different model types.

        Args:
            model_type (str): The type of the model (e.g., 'sense').
            messages (list): The list of messages to be sent to the model.
            gen_params (dict): The generation parameters.
            json_mode (bool): Flag to determine if the response format should be JSON.

        Returns:
            tuple: A tuple containing the header and the request data.
        """
        # Copy generation parameters to avoid modifying the original dictionary
        gen_params = gen_params.copy()

        # Hold out 100 tokens due to potential errors in token calculation
        max_tokens = min(gen_params.pop('max_new_tokens'), 4096)
        if max_tokens <= 0:
            return '', ''

        # Initialize the header
        header = {
            'content-type': 'application/json',
        }

        # Common parameters processing
        gen_params['max_tokens'] = max_tokens
        if 'stop_words' in gen_params:
            gen_params['stop'] = gen_params.pop('stop_words')
        if 'repetition_penalty' in gen_params:
            gen_params['frequency_penalty'] = gen_params.pop(
                'repetition_penalty')

        # Model-specific processing
        data = {}
        if model_type.lower().startswith('sense'):
            gen_params.pop('skip_special_tokens', None)
            gen_params.pop('session_id', None)
            data = {
                'model': model_type,
                'messages': messages,
                'n': 1,
                **gen_params
            }
            if json_mode:
                data['response_format'] = {'type': 'json_object'}
        else:
            raise NotImplementedError(
                f'Model type {model_type} is not supported')

        return header, data

    def tokenize(self, prompt: str) -> list:
        """Tokenize the input prompt.

        Args:
            prompt (str): Input string.

        Returns:
            list: token ids
        """
        import tiktoken

        self.tiktoken = tiktoken
        enc = self.tiktoken.encoding_for_model('gpt-4o')
        return enc.encode(prompt)



from itertools import count
from typing import Tuple

from lagent.schema import AgentMessage


class Hook:

    def before_agent(
        self,
        agent,
        message: Tuple[AgentMessage],
        session_id: int,
    ):
        pass

    def after_agent(
        self,
        agent,
        message: AgentMessage,
        session_id: int,
    ):
        pass

    def before_action(
        self,
        executor,
        message: AgentMessage,
        session_id: int,
    ):
        pass

    def after_action(
        self,
        executor,
        message: AgentMessage,
        session_id: int,
    ):
        pass


class RemovableHandle:
    _id_iter = count(0)

    def __init__(self, hooks_dict):
        self.hooks_dict = hooks_dict
        self.id = next(self._id_iter)

    def remove(self):
        del self.hooks_dict[self.id]



from .http_serve import AgentAPIServer, AsyncHTTPAgentClient, AsyncHTTPAgentServer, HTTPAgentClient, HTTPAgentServer
from .ray_serve import AgentRayActor, AsyncAgentRayActor

__all__ = [
    'AsyncAgentRayActor', 'AgentRayActor', 'HTTPAgentServer',
    'HTTPAgentClient', 'AsyncHTTPAgentServer', 'AsyncHTTPAgentClient',
    'AgentAPIServer'
]



# flake8: noqa: E501
import asyncio
import base64
import io
import json
import logging
import os
import queue
import re
import signal
import sys
import tempfile
import traceback
import uuid
from typing import Optional, Tuple, Type

from jupyter_client import AsyncKernelClient, AsyncKernelManager, AsyncMultiKernelManager
from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed

from lagent.actions.base_action import AsyncActionMixin, BaseAction, tool_api
from lagent.actions.parser import BaseParser, JsonParser
from lagent.schema import ActionReturn, ActionStatusCode

logger = logging.getLogger(__name__)

START_CODE = """
def input(*args, **kwargs):
    raise NotImplementedError('Python input() function is disabled.')

get_ipython().system = lambda *args: print('Assume we have this package, ! is disabled!')
{}
"""  # noqa


class TimeoutError(Exception):
    pass


class KernelDeath(Exception):
    pass


async def async_run_code(
    km: AsyncKernelManager,
    code,
    *,
    interrupt_after=30,
    iopub_timeout=40,
    wait_for_ready_timeout=60,
    shutdown_kernel=True,
):
    assert iopub_timeout > interrupt_after
    try:

        async def get_iopub_msg_with_death_detection(kc: AsyncKernelClient,
                                                     *,
                                                     timeout=None):
            loop = asyncio.get_running_loop()
            dead_fut = loop.create_future()

            def restarting():
                assert (
                    False
                ), "Restart shouldn't happen because config.KernelRestarter.restart_limit is expected to be set to 0"

            def dead():
                logger.info("Kernel has died, will NOT restart")
                dead_fut.set_result(None)

            msg_task = asyncio.create_task(kc.get_iopub_msg(timeout=timeout))
            km.add_restart_callback(restarting, "restart")
            km.add_restart_callback(dead, "dead")
            try:
                done, _ = await asyncio.wait(
                    [dead_fut, msg_task], return_when=asyncio.FIRST_COMPLETED)
                if dead_fut in done:
                    raise KernelDeath()
                assert msg_task in done
                return await msg_task
            finally:
                msg_task.cancel()
                km.remove_restart_callback(restarting, "restart")
                km.remove_restart_callback(dead, "dead")

        async def send_interrupt():
            await asyncio.sleep(interrupt_after)
            logger.info("Sending interrupt to kernel")
            await km.interrupt_kernel()

        @retry(
            retry=retry_if_result(lambda ret: ret[-1].strip() in [
                'KeyboardInterrupt',
                f"Kernel didn't respond in {wait_for_ready_timeout} seconds",
            ] if isinstance(ret, tuple) else False),
            stop=stop_after_attempt(3),
            wait=wait_fixed(1),
            retry_error_callback=lambda state: state.outcome.result())
        async def run():
            execute_result = None
            error_traceback = None
            stream_text_list = []
            kc = km.client()
            assert isinstance(kc, AsyncKernelClient)
            kc.start_channels()
            try:
                await kc.wait_for_ready(timeout=wait_for_ready_timeout)
                msg_id = kc.execute(code)
                while True:
                    message = await get_iopub_msg_with_death_detection(
                        kc, timeout=iopub_timeout)
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(
                            json.dumps(message, indent=2, default=str))
                    assert message["parent_header"]["msg_id"] == msg_id
                    msg_type = message["msg_type"]
                    if msg_type == "status":
                        if message["content"]["execution_state"] == "idle":
                            break
                    elif msg_type == "stream":
                        stream_name = message["content"]["name"]
                        stream_text = message["content"]["text"]
                        stream_text_list.append(stream_text)
                    elif msg_type == "execute_result":
                        execute_result = message["content"]["data"]
                    elif msg_type == "error":
                        error_traceback_lines = message["content"]["traceback"]
                        error_traceback = "\n".join(error_traceback_lines)
                    elif msg_type == "execute_input":
                        pass
                    else:
                        assert False, f"Unknown message_type: {msg_type}"
            finally:
                kc.stop_channels()
            return execute_result, error_traceback, "".join(stream_text_list)

        if interrupt_after:
            run_task = asyncio.create_task(run())
            send_interrupt_task = asyncio.create_task(send_interrupt())
            done, _ = await asyncio.wait([run_task, send_interrupt_task],
                                         return_when=asyncio.FIRST_COMPLETED)
            if run_task in done:
                send_interrupt_task.cancel()
            else:
                assert send_interrupt_task in done
            result = await run_task
        else:
            result = await run()
        return result
    finally:
        if shutdown_kernel:
            await km.shutdown_kernel()


class IPythonInterpreter(BaseAction):
    """A IPython executor that can execute Python scripts in a jupyter manner.

    Args:
        timeout (int): Upper bound of waiting time for Python script execution.
            Defaults to 20.
        user_data_dir (str, optional): Specified the user data directory for files
            loading. If set to `ENV`, use `USER_DATA_DIR` environment variable.
            Defaults to `ENV`.
        work_dir (str, optional): Specify which directory to save output images to.
            Defaults to ``'./work_dir/tmp_dir'``.
        description (dict): The description of the action. Defaults to ``None``.
        parser (Type[BaseParser]): The parser class to process the
            action's inputs and outputs. Defaults to :class:`JsonParser`.
    """

    _KERNEL_CLIENTS = {}

    def __init__(
        self,
        timeout: int = 20,
        user_data_dir: str = 'ENV',
        work_dir='./work_dir/tmp_dir',
        description: Optional[dict] = None,
        parser: Type[BaseParser] = JsonParser,
    ):
        super().__init__(description, parser)

        self.timeout = timeout
        if user_data_dir == 'ENV':
            user_data_dir = os.environ.get('USER_DATA_DIR', '')

        if user_data_dir:
            user_data_dir = os.path.dirname(user_data_dir)
            user_data_dir = f"import os\nos.chdir('{user_data_dir}')"
        self.user_data_dir = user_data_dir
        self._initialized = False
        self.work_dir = work_dir
        if not os.path.exists(self.work_dir):
            os.makedirs(self.work_dir, exist_ok=True)

    @staticmethod
    def start_kernel():
        from jupyter_client import KernelManager

        # start the kernel and manager
        km = KernelManager()
        km.start_kernel()
        kc = km.client()
        return km, kc

    def initialize(self):
        if self._initialized:
            return
        pid = os.getpid()
        if pid not in self._KERNEL_CLIENTS:
            self._KERNEL_CLIENTS[pid] = self.start_kernel()
        self.kernel_manager, self.kernel_client = self._KERNEL_CLIENTS[pid]
        self._initialized = True
        self._call(START_CODE.format(self.user_data_dir), None)

    def reset(self):
        if not self._initialized:
            self.initialize()
        else:
            code = "get_ipython().run_line_magic('reset', '-f')\n" + \
                START_CODE.format(self.user_data_dir)
            self._call(code, None)

    def _call(self,
              command: str,
              timeout: Optional[int] = None) -> Tuple[str, bool]:
        self.initialize()
        command = extract_code(command)

        # check previous remaining result
        while True:
            try:
                msg = self.kernel_client.get_iopub_msg(timeout=5)
                msg_type = msg['msg_type']
                if msg_type == 'status':
                    if msg['content'].get('execution_state') == 'idle':
                        break
            except queue.Empty:
                # assume no result
                break

        self.kernel_client.execute(command)

        def _inner_call():
            result = ''
            images = []
            succeed = True
            image_idx = 0

            while True:
                text = ''
                image = ''
                finished = False
                msg_type = 'error'
                try:
                    msg = self.kernel_client.get_iopub_msg(timeout=20)
                    msg_type = msg['msg_type']
                    if msg_type == 'status':
                        if msg['content'].get('execution_state') == 'idle':
                            finished = True
                    elif msg_type == 'execute_result':
                        text = msg['content']['data'].get('text/plain', '')
                        if 'image/png' in msg['content']['data']:
                            image_b64 = msg['content']['data']['image/png']
                            image_url = publish_image_to_local(
                                image_b64, self.work_dir)
                            image_idx += 1
                            image = '![fig-%03d](%s)' % (image_idx, image_url)

                    elif msg_type == 'display_data':
                        if 'image/png' in msg['content']['data']:
                            image_b64 = msg['content']['data']['image/png']
                            image_url = publish_image_to_local(
                                image_b64, self.work_dir)
                            image_idx += 1
                            image = '![fig-%03d](%s)' % (image_idx, image_url)

                        else:
                            text = msg['content']['data'].get('text/plain', '')
                    elif msg_type == 'stream':
                        msg_type = msg['content']['name']  # stdout, stderr
                        text = msg['content']['text']
                    elif msg_type == 'error':
                        succeed = False
                        text = escape_ansi('\n'.join(
                            msg['content']['traceback']))
                        if 'M6_CODE_INTERPRETER_TIMEOUT' in text:
                            text = f'Timeout. No response after {timeout} seconds.'  # noqa
                except queue.Empty:
                    # stop current task in case break next input.
                    self.kernel_manager.interrupt_kernel()
                    succeed = False
                    text = f'Timeout. No response after {timeout} seconds.'
                    finished = True
                except Exception:
                    succeed = False
                    msg = ''.join(traceback.format_exception(*sys.exc_info()))
                    # text = 'The code interpreter encountered an unexpected error.'  # noqa
                    text = msg
                    logging.warning(msg)
                    finished = True
                if text:
                    # result += f'\n\n{msg_type}:\n\n```\n{text}\n```'
                    result += f'{text}'

                if image:
                    images.append(image_url)
                if finished:
                    return succeed, dict(text=result, image=images)

        try:
            if timeout:

                def handler(signum, frame):
                    raise TimeoutError()

                signal.signal(signal.SIGALRM, handler)
                signal.alarm(timeout)
            succeed, result = _inner_call()
        except TimeoutError:
            succeed = False
            text = 'The code interpreter encountered an unexpected error.'
            result = f'\n\nerror:\n\n```\n{text}\n```'
        finally:
            if timeout:
                signal.alarm(0)

        # result = result.strip('\n')
        return succeed, result

    @tool_api
    def run(self, command: str, timeout: Optional[int] = None) -> ActionReturn:
        r"""When you send a message containing Python code to python, it will be executed in a stateful Jupyter notebook environment. python will respond with the output of the execution or time out after 60.0 seconds. The drive at '/mnt/data' can be used to save and persist user files. Internet access for this session is disabled. Do not make external web requests or API calls as they will fail.

        Args:
            command (:class:`str`): Python code
            timeout (:class:`Optional[int]`): Upper bound of waiting time for Python script execution.
        """
        tool_return = ActionReturn(url=None, args=None, type=self.name)
        tool_return.args = dict(text=command)
        succeed, result = self._call(command, timeout)
        if succeed:
            text = result['text']
            image = result.get('image', [])
            resp = [dict(type='text', content=text)]
            if image:
                resp.extend([dict(type='image', content=im) for im in image])
            tool_return.result = resp
            # tool_return.result = dict(
            #     text=result['text'], image=result.get('image', [])[0])
            tool_return.state = ActionStatusCode.SUCCESS
        else:
            tool_return.errmsg = result.get('text', '') if isinstance(
                result, dict) else result
            tool_return.state = ActionStatusCode.API_ERROR
        return tool_return


class AsyncIPythonInterpreter(AsyncActionMixin, IPythonInterpreter):
    """A IPython executor that can execute Python scripts in a jupyter manner.

    Args:
        timeout (int): Upper bound of waiting time for Python script execution.
            Defaults to 20.
        user_data_dir (str, optional): Specified the user data directory for files
            loading. If set to `ENV`, use `USER_DATA_DIR` environment variable.
            Defaults to `ENV`.
        work_dir (str, optional): Specify which directory to save output images to.
            Defaults to ``'./work_dir/tmp_dir'``.
        description (dict): The description of the action. Defaults to ``None``.
        parser (Type[BaseParser]): The parser class to process the
            action's inputs and outputs. Defaults to :class:`JsonParser`.
    """

    _UNBOUND_KERNEL_CLIENTS = asyncio.Queue()

    def __init__(
        self,
        timeout: int = 20,
        user_data_dir: str = 'ENV',
        work_dir=os.path.join(tempfile.gettempdir(), 'tmp_dir'),
        max_kernels: Optional[int] = None,
        reuse_kernel: bool = True,
        startup_rate: bool = 32,
        connection_dir: str = tempfile.gettempdir(),
        description: Optional[dict] = None,
        parser: Type[BaseParser] = JsonParser,
    ):
        super().__init__(timeout, user_data_dir, work_dir, description, parser)
        from traitlets.config import Config

        c = Config()
        c.KernelManager.transport = 'ipc'
        self._amkm = AsyncMultiKernelManager(
            config=c, connection_dir=connection_dir)
        self._max_kernels = max_kernels
        self._reuse_kernel = reuse_kernel
        self._sem = asyncio.Semaphore(startup_rate)
        self._lock = asyncio.Lock()

    async def initialize(self, session_id: str):
        session_id = str(session_id)
        while True:
            if session_id in self._KERNEL_CLIENTS:
                return self._KERNEL_CLIENTS[session_id]
            if self._reuse_kernel and not self._UNBOUND_KERNEL_CLIENTS.empty():
                self._KERNEL_CLIENTS[
                    session_id] = await self._UNBOUND_KERNEL_CLIENTS.get()
                return self._KERNEL_CLIENTS[session_id]
            async with self._sem:
                if self._max_kernels is None or len(
                        self._KERNEL_CLIENTS
                ) + self._UNBOUND_KERNEL_CLIENTS.qsize() < self._max_kernels:
                    kernel_id = None
                    try:
                        kernel_id = await self._amkm.start_kernel()
                        kernel = self._amkm.get_kernel(kernel_id)
                        client = kernel.client()
                        _, error_stacktrace, stream_text = await async_run_code(
                            kernel,
                            START_CODE.format(self.user_data_dir),
                            shutdown_kernel=False)
                        # check if the output of START_CODE meets expectations
                        if not (error_stacktrace is None
                                and stream_text == ''):
                            raise RuntimeError
                    except Exception as e:
                        print(f'Starting kernel error: {e}')
                        if kernel_id:
                            await self._amkm.shutdown_kernel(kernel_id)
                            self._amkm.remove_kernel(kernel_id)
                        await asyncio.sleep(1)
                        continue
                    if self._max_kernels is None:
                        self._KERNEL_CLIENTS[session_id] = (kernel_id, kernel,
                                                            client)
                        return kernel_id, kernel, client
                    async with self._lock:
                        if len(self._KERNEL_CLIENTS
                               ) + self._UNBOUND_KERNEL_CLIENTS.qsize(
                               ) < self._max_kernels:
                            self._KERNEL_CLIENTS[session_id] = (kernel_id,
                                                                kernel, client)
                            return kernel_id, kernel, client
                    await self._amkm.shutdown_kernel(kernel_id)
                    self._amkm.remove_kernel(kernel_id)
            await asyncio.sleep(1)

    async def reset(self, session_id: str):
        session_id = str(session_id)
        if session_id not in self._KERNEL_CLIENTS:
            return
        _, kernel, _ = self._KERNEL_CLIENTS[session_id]
        code = "get_ipython().run_line_magic('reset', '-f')\n" + \
            START_CODE.format(self.user_data_dir)
        await async_run_code(kernel, code, shutdown_kernel=False)

    async def shutdown(self, session_id: str):
        session_id = str(session_id)
        if session_id in self._KERNEL_CLIENTS:
            kernel_id, _, _ = self._KERNEL_CLIENTS.get(session_id)
            await self._amkm.shutdown_kernel(kernel_id)
            self._amkm.remove_kernel(kernel_id)
            del self._KERNEL_CLIENTS[session_id]

    async def close_session(self, session_id: str):
        session_id = str(session_id)
        if self._reuse_kernel:
            if session_id in self._KERNEL_CLIENTS:
                await self.reset(session_id)
                await self._UNBOUND_KERNEL_CLIENTS.put(
                    self._KERNEL_CLIENTS.pop(session_id))
        else:
            await self.shutdown(session_id)

    async def _call(self, command, timeout=None, session_id=None):
        _, kernel, _ = await self.initialize(str(session_id))
        result = await async_run_code(
            kernel,
            extract_code(command),
            interrupt_after=timeout or self.timeout,
            shutdown_kernel=False)
        execute_result, error_stacktrace, stream_text = result
        if error_stacktrace is not None:
            ret = re.sub('^-*\n', '', escape_ansi(error_stacktrace))
            if ret.endswith('KeyboardInterrupt: '):
                ret = 'The code interpreter encountered a timeout error.'
            status, ret = False, ret.strip()
        elif execute_result is not None:
            status, ret = True, dict(text=execute_result.get('text/plain', ''))
        else:
            status, ret = True, dict(text=stream_text.strip())
        return status, ret

    @tool_api
    async def run(self,
                  command: str,
                  timeout: Optional[int] = None,
                  session_id: Optional[str] = None) -> ActionReturn:
        r"""When you send a message containing Python code to python, it will be executed in a stateful Jupyter notebook environment. python will respond with the output of the execution or time out after 60.0 seconds. The drive at '/mnt/data' can be used to save and persist user files. Internet access for this session is disabled. Do not make external web requests or API calls as they will fail.

        Args:
            command (:class:`str`): Python code
            timeout (:class:`Optional[int]`): Upper bound of waiting time for Python script execution.
        """
        tool_return = ActionReturn(url=None, args=None, type=self.name)
        tool_return.args = dict(text=command)
        succeed, result = await self._call(command, timeout, session_id)
        if succeed:
            text = result['text']
            image = result.get('image', [])
            resp = [dict(type='text', content=text)]
            if image:
                resp.extend([dict(type='image', content=im) for im in image])
            tool_return.result = resp
            # tool_return.result = dict(
            #     text=result['text'], image=result.get('image', [])[0])
            tool_return.state = ActionStatusCode.SUCCESS
        else:
            tool_return.errmsg = result.get('text', '') if isinstance(
                result, dict) else result
            tool_return.state = ActionStatusCode.API_ERROR
        return tool_return


def extract_code(text):
    import json5

    # Match triple backtick blocks first
    triple_match = re.search(r'```[^\n]*\n(.+?)```', text, re.DOTALL)
    # Match single backtick blocks second
    single_match = re.search(r'`([^`]*)`', text, re.DOTALL)
    if triple_match:
        text = triple_match.group(1)
    elif single_match:
        text = single_match.group(1)
    else:
        try:
            text = json5.loads(text)['code']
        except Exception:
            pass
    # If no code blocks found, return original text
    return text


def escape_ansi(line):
    ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', line)


def publish_image_to_local(image_base64: str, work_dir='./work_dir/tmp_dir'):
    import PIL.Image
    image_file = str(uuid.uuid4()) + '.png'
    local_image_file = os.path.join(work_dir, image_file)

    png_bytes = base64.b64decode(image_base64)
    assert isinstance(png_bytes, bytes)
    bytes_io = io.BytesIO(png_bytes)
    PIL.Image.open(bytes_io).save(local_image_file, 'png')

    return local_image_file


# local test for code interpreter
def get_multiline_input(hint):
    print(hint)
    print('// Press ENTER to make a new line. Press CTRL-D to end input.')
    lines = []
    while True:
        try:
            line = input()
        except EOFError:  # CTRL-D
            break
        lines.append(line)
    print('// Input received.')
    if lines:
        return '\n'.join(lines)
    else:
        return ''


if __name__ == '__main__':
    code_interpreter = IPythonInterpreter()
    while True:
        print(code_interpreter(get_multiline_input('Enter python code:')))
